; Tools for writing games based on "actors" with their own positions,
; velocities, and tags.

(define (new-VelocityEngine xform)
  (define timer (new-SoElapsedTime))
  (define calc (new-SoCalculator))
  (-> (-> calc '_a) 'connectFrom (-> timer 'timeOut))

  (define velocity (-> calc 'A))
  (-> velocity 'setValue #(0 0 0))

  (define offset (-> calc 'B))
  (define position (-> xform 'translation))
  (-> offset 'setValue (-> (-> position 'getValue) 'getValue))
  (-> (-> calc 'expression) 'setValue "oA = a*A + B")
  (-> position 'connectFrom (-> calc 'oA))

  (lambda (message)
    (cond ((eq? message 'velocity)
	   (lambda (self) velocity))

	  ((eq? message 'offset)
	   (lambda (self) offset))

	  ((eq? message 'position)
	   (lambda (self) position))

	  ((eq? message 'getPosition)
	   (lambda (self) (-> (-> position 'getValue) 'getValue)))

	  ((eq? message 'setPosition)
	   (lambda (self p) 
	     (-> offset 'setValue p)
	     (-> (-> timer 'reset) 'setValue)))

	  ((eq? message 'getVelocity)
	   (lambda (self) (-> (-> velocity 'operator-brackets 0) 'getValue)))

	  ((eq? message 'setVelocity)
	   (lambda (self v)
	     (send self 'setPosition (send self 'getPosition))
	     (-> velocity 'setValue v)
	     ))

	  ((eq? message 'start)
	   (lambda (self)
	     (-> (-> timer 'on) 'setValue 1)))

	  ((eq? message 'die)
	   (lambda (self)
	     (-> (-> calc '_a) 'disconnect)
	     (-> position 'disconnect)
	     ))

	  (else (no-method message))
	  )))

(define (filter pred lst)
  (cond ((null? lst) '())
	((pred (car lst))
	 (cons (car lst) (filter pred (cdr lst))))
	(else (filter pred (cdr lst)))))

(define (fold1 op lst)
  (cond ((null? lst) #f)
	((pair? (cdr lst))
	 (op (car lst) (fold1 op (cdr lst))))
	(else (car lst))))

(define (new-TagTable)
  (define table '())
  (lambda (message)
    (cond ((eq? message 'withTag)
	   (lambda (self tag) 
	     (define pair (assq tag table))
	     (if pair
		 (cdr pair)
		 '())))

	  ((eq? message 'addTag)
	   (lambda (self tag obj)
	     (define pair (assq tag table))
	     (if pair
		 (set-cdr! pair (cons obj (cdr pair)))
		 (set! table (cons (list tag obj) table)))))
	  
	  ((eq? message 'removeTag)
	   (lambda (self tag obj)
	     (define pair (assq tag table))
	     (if pair
		 (set-cdr! pair (filter (lambda (o) (not (eq? o obj))) 
					(cdr pair))))
	     ))

	  (else (no-method message))
	  )))
(define *actor-tags* (new-TagTable))

(define (new-Actor root shape pos)
  (define my-root (new-SoSeparator))
  (-> my-root 'ref)
  (define trans (new-SoTransform))
  (-> (-> trans 'translation) 'setValue 
      (vector-ref pos 0) (vector-ref pos 1) (vector-ref pos 2))
  (addChildren my-root trans shape)
  (addChildren root my-root)
  (define ve (new-VelocityEngine trans))
  (define bbox #f)
  (define tags '())
  (define dead #f)

  (define (self message)
    (cond (dead
	   (lambda (self . args)
	     (display "error: called dead actor\n")))

	  ((eq? message 'removeAllTags)
	   (lambda (self)
	     (for-each (lambda (tag) (send *actor-tags* 'removeTag tag self))
		       tags)))

	  ((eq? message 'die)
	   (lambda (self) 
	     (if (not (null? positionHooks))
		 (-> position-sensor 'detach))
	     (send ve 'die)
	     (send self 'removeAllTags)
	     (set! dead #t)
	     (-> root 'removeChild my-root)
	     #f))

	  ((eq? message 'shape)
	   (lambda (self) shape))

	  ((eq? message 'bbox)
	   ; returns an SbBox3f
	   (lambda (self viewer)
	     (cond ((not bbox)
		    (define viewportRegion (-> viewer 'getViewportRegion))
		    (define action (new-SoGetBoundingBoxAction viewportRegion))
		    (-> action 'apply my-root)
		    (define real-bbox (-> action 'getBoundingBox))

		    (define min (-> real-bbox 'getMin))
		    (define max (-> real-bbox 'getMax))
		    (define c (-> (send self 'position) 'getValue))
		    (set! min (-> min 'operator- c))
		    (set! max (-> max 'operator- c))
		    (set! bbox (new-SbBox3f min max))
		    )
		   (else
		    (define min (-> bbox 'getMin))
		    (define max (-> bbox 'getMax))
		    (define c (-> (send self 'position) 'getValue))
		    (set! min (-> min 'operator+ c))
		    (set! max (-> max 'operator+ c))
		    (define real-bbox (new-SbBox3f min max))
		    ))
	     real-bbox))

	  ((eq? message 'addPositionHook)
	   (lambda (self hook)
	     (if (null? positionHooks)
		 (-> position-sensor 'attach (send self 'position)))
	     (set! positionHooks (cons (list hook self) positionHooks))))

	  ((eq? message 'removePositionHook)
	   (lambda (self hook)
	     (set! positionHooks 
		   (filter (lambda (hook+args) 
			     (not (eq? (car hook+args) hook))) 
			   positionHooks))))

	  ((eq? message 'addCollisionHandler)
	   (lambda (self viewer tags proc)
	     (define (positionHook self)
	       (for-each
		(lambda (other)
		  (cond ((and (not (eq? self other))
			      (actor-collision? self other viewer))
			 (proc self other))
			))
		(fold1 append 
		       (map (lambda (tag) (send *actor-tags* 'withTag tag)) 
			    tags))
	       ))
	     (send self 'addPositionHook positionHook)))

	  ((eq? message 'addBoundsCheck)
	   (lambda (self box faces)
	     (define (positionHook self)
	       (actor-bounds-check self box faces))
	     (send self 'addPositionHook positionHook)))

	  ((eq? message 'tags)
	   (lambda (self) tags))

	  ((eq? message 'addTag)
	   (lambda (self tag) 
	     (send *actor-tags* 'addTag tag self)
	     (set! tags (cons tag tags))))

	  (else (ve message))
	  ))

  (define positionHooks '())
  (define (sensor-cb user-data sensor)
    (for-each (lambda (hook+args) (apply (car hook+args) (cdr hook+args)))
	      positionHooks))
  (define position-sensor (sensor new-SoFieldSensor sensor-cb))

  ;(send self 'addTag 'any)
  self)

; returns #t if actor1 and actor2 have collided
(define (actor-collision? actor1 actor2 viewer)
  (= 1 (-> (send actor1 'bbox viewer) 'intersect (send actor2 'bbox viewer))))

; moves actor1 and actor2 apart so that they no longer intersect
(define (actor-separate actor1 actor2 viewer)
  (define bbox1 (send actor1 'bbox viewer))
  (define min1 (-> (-> bbox1 'getMin) 'getValue))
  (define max1 (-> (-> bbox1 'getMax) 'getValue))
  (define p1 (send actor1 'getPosition))
  (define v1 (send actor1 'getVelocity))

  (define bbox2 (send actor2 'bbox viewer))
  (define min2 (-> (-> bbox2 'getMin) 'getValue))
  (define max2 (-> (-> bbox2 'getMax) 'getValue))
  (define p2 (send actor2 'getPosition))
  (define v2 (send actor2 'getVelocity))
  
  (define t 100)
  (for-each
   (lambda (d)
     ; compute combined velocity
     (define v (- (vector-ref v1 d) (vector-ref v2 d)))
     ; prevent divide by zero
     (if (= v 0) (set! v 1e-10))
     ; compute gap
     (if (> v 0)
	 (define gap (- (vector-ref max1 d) (vector-ref min2 d)))
	 (define gap (- (vector-ref max2 d) (vector-ref min1 d))))
     ; make sure actors really get separated
     (set! gap (+ gap 1e-2))
     ; compute time to cross the gap
     (define gt (/ gap (abs v)))
     ; send t to min(t, gt)
     (if (< gt t)
	 (set! t gt))
     )
   '(0 1 2))

  ; move each actor back in time by t
  (vector-add p1 (vector-scale (- t) v1))
  (vector-add p2 (vector-scale (- t) v2))

  (send actor1 'setPosition p1)
  (send actor2 'setPosition p2)

  ;(if (actor-collision? actor1 actor2 viewer)
  ;    (begin (display "foo\n")
  ;	     (actor-separate actor1 actor2 viewer)))
  )

; makes sure actor is within the box (an SbBox3f)
; faces is vector of 6 numbers which determine face behavior
; 0 for stop, >0 for wrap, <0 for bounce
(define (actor-bounds-check actor box faces)
  (define updated #f)
  (define p (send actor 'getPosition))
  (define v (send actor 'getVelocity))
  (define min (-> (-> box 'getMin) 'getValue))
  (define max (-> (-> box 'getMax) 'getValue))

  (for-each 
   (lambda (face)
     (define d (quotient face 2))
     (if (= 0 (modulo face 2))
	 (if (< (vector-ref p d) (vector-ref min d))
	     (begin
	       (cond ((= 0 (vector-ref faces face))
		      ; stop
		      (vector-set! p d (vector-ref min d))
		      (vector-set! v d 0))
		     
		     ((< 0 (vector-ref faces face))
		      ; wrap
		      (vector-set! p d (+ (vector-ref p d) 
					  (- (vector-ref min d))
					  (vector-ref max d))))
		     
		     (else
		      ; bounce
		      (vector-set! p d (- (* 2 (vector-ref min d)) 
					  (vector-ref p d)))
		      (vector-set! v d (* (vector-ref faces face) 
					  (vector-ref v d)))))
	       (set! updated #t)))
	 (if (> (vector-ref p d) (vector-ref max d))
	     (begin
	       (cond ((= 0 (vector-ref faces face))
		      ; stop
		      (vector-set! p d (vector-ref max d))
		      (vector-set! v d 0))
		     
		     ((< 0 (vector-ref faces face))
		      ; wrap
		      (vector-set! p d (+ (vector-ref p d)
					  (vector-ref min d) 
					  (- (vector-ref max d)))))
		     
		     (else
		      ; bounce
		      (vector-set! p d (- (* 2 (vector-ref max d)) 
					  (vector-ref p d)))
		      (vector-set! v d (* (vector-ref faces face)
					  (vector-ref v d)))))
	       (set! updated #t)))
	 ))
   '(0 1 2 3 4 5))

  (if updated
      (begin 
	(send actor 'setPosition p)
	(send actor 'setVelocity v)))
  )

; represent v in the given orthonormal basis
; axes are SbVec3f while v is vector
(define (rebase-SbVec3f axis1 axis2 axis3 v)
  (define vec (new-SbVec3f v))
  (define p1 (-> vec 'dot axis1))
  (define p2 (-> vec 'dot axis2))
  (define p3 (-> vec 'dot axis3))
  (vector p1 p2 p3))

; handle an elastic collision between actor1 and actor2
; mass holds the inverse masses of the actors
(define (elastic-collision actor1 actor2 . mass)
  (if (not (pair? mass)) (set! mass '(1 1)))
  (define m1 (car mass))
  (define m2 (cadr mass))
  (define v1 (send actor1 'getVelocity))
  (define v2 (send actor2 'getVelocity))
  
  ; contact axis for spherical objects
  (define axis1 (-> (-> (send actor1 'position) 'getValue) 'operator- 
		        (-> (send actor2 'position) 'getValue)))
  (-> axis1 'normalize)

  ; orthogonal axis
  (define axis2 (new-SbVec3f (-> axis1 'operator-brackets 1)
			     (- (-> axis1 'operator-brackets 0))
			     0))
  (-> axis2 'normalize)

  ; another orthogonal axis
  (define axis3 (-> axis2 'cross axis1))
  (-> axis3 'normalize)

  ; switch to new basis
  (set! v1 (rebase-SbVec3f axis1 axis2 axis3 v1))
  (set! v2 (rebase-SbVec3f axis1 axis2 axis3 v2))

  ; compute new velocities along the contact axis
  (define mt (+ m2 m1))
  (define md (- m2 m1))
  (define u1 (vector-ref v1 0))
  (define u2 (vector-ref v2 0))
  (vector-set! v1 0 (/ (+ (* 2 m1 u2) (* md u1)) mt))
  (vector-set! v2 0 (/ (- (* 2 m2 u1) (* md u2)) mt))

  ; back to old basis
  (set! v1 (rebase-SbVec3f axis1 axis2 axis3 v1))
  (set! v2 (rebase-SbVec3f axis1 axis2 axis3 v2))

  (actor-separate actor1 actor2 viewer)
  (send actor1 'setVelocity v1)
  (send actor2 'setVelocity v2))

; crude approximation to the above
(define (elastic-collision-simple actor1 actor2)
  (define v1 (send actor1 'getVelocity))
  (define v2 (send actor2 'getVelocity))
  
  (actor-separate actor1 actor2 viewer)
  (send actor1 'setVelocity v2)
  (send actor2 'setVelocity v1))

(define *keyboard-active* #t)

(define (keyboard-handler root proc)
  (define (key-event-cb user-data event-node)
    (let ((event (-> event-node 'getEvent)))
      (proc event)
      ;; tell the callback node that we have handled the event.
      (-> event-node 'setHandled)))

  (define event-callback (new-SoEventCallback))
  (-> root 'addChild event-callback)
  (-> event-callback 'addEventCallback (SoKeyboardEvent::getClassTypeId)
      (get-scheme-event-callback-cb)
      (void-cast (callback-info key-event-cb)))
  )
